"""Climax configuration file."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import datetime

# externals
import torch
import numpy as np

# locals
from climax.core.constants import (PREDICTANDS, ERA5_P_VARIABLES,
                                   ERA5_S_VARIABLES)
from climax.core.models import SegNet, USegNet, StridedSegNet
from climax.core.loss import (BernoulliGammaLoss, MSELoss, L1Loss,
                              BernoulliWeibullLoss)

# -----------------------------------------------------------------------------
# ERA5 downscaling configuration ----------------------------------------------
# -----------------------------------------------------------------------------

# ERA5 predictor variables on pressure levels
ERA5_P_PREDICTORS = ['geopotential', 'temperature', 'u_component_of_wind',
                     'v_component_of_wind', 'specific_humidity']
# ERA5_P_PREDICTORS = []
assert all([var in ERA5_P_VARIABLES for var in ERA5_P_PREDICTORS])

# ERA5 predictor variables on single levels
# ERA5_S_PREDICTORS = ['mean_sea_level_pressure', 'orography', '2m_temperature']
# ERA5_S_PREDICTORS = ['mean_sea_level_pressure']
ERA5_S_PREDICTORS = ['surface_pressure']
# ERA5_S_PREDICTORS = ['total_precipitation']
assert all([var in ERA5_S_VARIABLES for var in ERA5_S_PREDICTORS])

# ERA5 predictor variables
ERA5_PREDICTORS = ERA5_P_PREDICTORS + ERA5_S_PREDICTORS

# ERA5 pressure levels
ERA5_PLEVELS = [500, 850]

# whether to train on observed time series or on standardized anomalies
# Anomaly = (time_series  - mean(time_series)) / (std(time_series))
ANOMALIES = False

# Dask chunk size for loading the training data
CHUNKS = {'time': 365}

# -----------------------------------------------------------------------------
# Auxiliary predictors --------------------------------------------------------
# -----------------------------------------------------------------------------

# include day of year as predictor
DOY = True

# use digital elevation model instead of model orography
DEM = True
if DEM:
    # remove model orography when using DEM
    if 'orography' in ERA5_S_PREDICTORS:
        ERA5_S_PREDICTORS.remove('orography')

# whether to use DEM slope and aspect as predictors
DEM_FEATURES = False

# -----------------------------------------------------------------------------
# Training, Validation, and Test datasets -------------------------------------
# -----------------------------------------------------------------------------

# calibration period: training and validation
CALIB_PERIOD = np.arange(
    datetime.datetime.strptime('1981-01-01', '%Y-%m-%d').date(),
    datetime.datetime.strptime('1991-01-01', '%Y-%m-%d').date())

# validation period: testing
VALID_PERIOD = np.arange(
    datetime.datetime.strptime('1991-01-01', '%Y-%m-%d').date(),
    datetime.datetime.strptime('2011-01-01', '%Y-%m-%d').date())

# entire reference period
REFERENCE_PERIOD = np.concatenate([CALIB_PERIOD, VALID_PERIOD], axis=0)

# stratify training/validation set for precipitation by number of wet days
STRATIFY = False

# size of the validation set w.r.t. the training set
# e.g., VALID_SIZE = 0.2 means: 80% of CALIB_PERIOD for training
#                               20% of CALIB_PERIOD for validation
VALID_SIZE = 0.2

# number of folds for training with KFold cross-validation
CV = 5

# number of bootstrapped model trainings
BOOTSTRAP = 10

# -----------------------------------------------------------------------------
# Observations ----------------------------------------------------------------
# -----------------------------------------------------------------------------

# target variable: check if target variable is valid
PREDICTAND = 'pr'
assert PREDICTAND in PREDICTANDS

# threshold  defining the minimum amount of precipitation (mm) for a wet day
WET_DAY_THRESHOLD = 1

# whether training a hyperparameter sensitivity analysis
SENSITIVITY = False

# -----------------------------------------------------------------------------
# Model training configuration ------------------------------------------------
# -----------------------------------------------------------------------------

# network architecture
NET = USegNet

# number of convolutional layers and corresponding filter sizes
FILTERS = [32, 64, 128, 256]

# loss function to optimize
# for temperature: L1Loss or MSELoss (NLL of normal distribution)
# for precipitation:
#     BernoulliGammaLoss (NLL of Bernoulli-Gamma distribution)
#     BernoulliWeibullLoss (NLL of Bernoulli-Weibull distribution)
# LOSS = L1Loss()
LOSS = MSELoss()
# LOSS = BernoulliGammaLoss(min_amount=1)
# LOSS = BernoulliWeibullLoss(min_amount=1)

# stochastic optimization algorithm
# OPTIM = torch.optim.SGD
OPTIM = torch.optim.Adam

# batch size: number of time steps processed by the net in each iteration
BATCH_SIZE = 16

# maximum learning rate determined from learning rate range test
if PREDICTAND is 'tasmin':
    if isinstance(LOSS, L1Loss):
        MAX_LR = 0.001 if OPTIM is torch.optim.Adam else 0.004
    if isinstance(LOSS, MSELoss):
        MAX_LR = 0.001 if OPTIM is torch.optim.Adam else 0.002

if PREDICTAND is 'tasmax':
    if isinstance(LOSS, L1Loss):
        MAX_LR = 0.001
    if isinstance(LOSS, MSELoss):
        MAX_LR = 0.001 if OPTIM is torch.optim.Adam else 0.004

if PREDICTAND is 'pr':
    if isinstance(LOSS, L1Loss):
        MAX_LR = 0.001
    if isinstance(LOSS, MSELoss):
        MAX_LR = 0.0004
    if isinstance(LOSS, BernoulliGammaLoss):
        # learning rates for supersampling task
        if not any(ERA5_P_PREDICTORS) and ERA5_S_PREDICTORS == 'pr':
            MAX_LR = 0.001
        else:
            MAX_LR = 0.0005 if OPTIM is torch.optim.Adam else 0.001

# base learning rate: MAX_LR / 4 (Smith L. (2017))
BASE_LR = MAX_LR / 4

# optimization parameters
OPTIM_PARAMS = {'lr': BASE_LR,
                'weight_decay': 0
                }
if OPTIM is torch.optim.SGD:
    OPTIM_PARAMS['momentum'] = 0.99  # SGD with momentum

# learning rate scheduler: CyclicLR policy
LR_SCHEDULER = None
# LR_SCHEDULER = torch.optim.lr_scheduler.CyclicLR
LR_SCHEDULER_PARAMS = {
    'base_lr': BASE_LR, 'max_lr': MAX_LR, 'mode': 'triangular',
    'cycle_momentum': True if OPTIM is torch.optim.SGD else False,
    'base_momentum': 0.9, 'max_momentum': 0.99, 'step_size_up': 400}

# whether to randomly shuffle time steps or to conserve time series for model
# training
SHUFFLE = True

# whether to normalize the training data to [0, 1]
NORM = True

# network training configuration
TRAIN_CONFIG = {
    'checkpoint_state': {},
    'epochs': 100,
    'save': True,
    'save_loaders': False,
    'early_stop': True,
    'patience': 10,
    'multi_gpu': True,
    'classification': False,
    'clip_gradients': False
    }

# whether to overwrite existing models
OVERWRITE = False

# split reference period into NYEARS intervals
# this limits the amount of data that is loaded into memory during prediction
# NOTE: if NYEARS=None, load entire reference period into memory at once
NYEARS = 5
